/* Hints for the author of the present program:
   1) Add useful comments to claim the aims of each function
      the name of which begins with "write_..."
   2) Eventually split this file in three parts:
      the first one with the main function and just a few other ones;
      the second with the functions allowing the handling (i.e. create,
      write, etc.) of the expansions;
      the third one with the functions allowing to perform algebraic
      operations on the expansions */

#include "common_def.h"


/* ### */
/* Useful constant parameters */
#define NDIM 1
#define NDOF (NDIM + 1)
#define K 2
#define KAM_NORMAL_ORDER 25
#define MAXPOLDEG 2
#define MAXMODANG (KAM_NORMAL_ORDER * K)
#define EPSILON 0.012


/* ### */
/* Global variables */
unsigned long int Jpol_table[NDIM+1][MAXPOLDEG+1];
unsigned long int Npol_table[NDIM+1][MAXPOLDEG+1];
unsigned long int Jtrig_table[NDOF+1][MAXMODANG+1];
unsigned long int Ntrig_table[NDOF+1][MAXMODANG+1];

/* List of the pointers to the first address of the expansion of the
   functions we need */
INTERVAL *ham[MAXPOLDEG+1];
INTERVAL *old_ham[MAXPOLDEG+1];
INTERVAL *average_ham[3];
INTERVAL *tmp_ham[2][MAXPOLDEG+1];
INTERVAL *gen_X;
INTERVAL *gen_csiq;
INTERVAL *gen_chi2;
INTERVAL *target_X;
INTERVAL *target_csiq;
INTERVAL *target_chi2;

/* Length of the expansions we need */
unsigned long int nterms_ham[MAXPOLDEG+1];
unsigned long int nterms_aver_ham[3];

/* Values of the frequency vector characterizing the invariant torus
   we are interested in */
INTERVAL omega[NDOF];


/* ### */
/* Function prototypes */

/* About the handling of the expansions */
void create_functions();
void def_init_ham();
void funct_norm(INTERVAL *fn_addr, unsigned long int nterms,
		unsigned short int pol_deg, short int min_harmonic_length,
		short int max_harmonic_length, INTERVAL *norm);
void print_nz_terms(INTERVAL *fn_addr, unsigned long int nterms,
		    unsigned short int pol_deg, char flag_average);
void write_norms_ham_terms(FILE *ofp);
void write_util_estimates_lie_series_chi2(FILE *ofp,
					  unsigned short int ord_gen);
void write_util_estimates_lie_series_csiq(FILE *ofp);
void write_util_estimates_lie_series_X(FILE *ofp,
				       unsigned short int ord_gen);


/* About the algebraic operations on the expansions */
void add_fn_to(INTERVAL *fn1_addr, unsigned long int nterms,
	       INTERVAL *fn2_addr);
void add_poisson_to(INTERVAL *fn1_addr, unsigned long int n1_terms,
		    unsigned short int pol1_deg, INTERVAL *fn2_addr,
		    unsigned long int n2_terms, unsigned short int pol2_deg,
		    INTERVAL coef_scal_mul, INTERVAL *target_fn_addr);
void add_poisson_transl_to(INTERVAL *gen_transl_addr, INTERVAL *fn2_addr,
			   unsigned long int n2_terms,
			   unsigned short int pol2_deg,
			   INTERVAL coef_scal_mul, INTERVAL *target_fn_addr);
void add_scalarmultiply_fn_to(INTERVAL *fn1_addr, unsigned long int nterms,
			      INTERVAL mul_coef, INTERVAL *fn2_addr);
void build_index_pol_table();
void build_index_trig_table();
void calc_inv_matr_C(INTERVAL *aver_quadr_fn_addr,
		     INTERVAL inv_matr_C[][NDIM]);
void can_transf_ham(INTERVAL *gen_fn_addr, unsigned short int ord_gen,
		    unsigned short int gen_pol_deg, char flag_transl);
void clean_function(INTERVAL *fn_addr, unsigned long int nterms);
unsigned long int code_hom_pol_ind(unsigned short int kpol[]);
unsigned long int code_trig_ind(signed short int ktrig[]);
void cramer(INTERVAL A[][NDIM], short int n, INTERVAL inv_A[][NDIM]);
void cut_0row_icolumn(INTERVAL matr[][NDIM], short int n,
		      short int i, INTERVAL reduced_matr[][NDIM]);
void cut_harmonics(INTERVAL *fn_addr, unsigned long int nterms,
		   unsigned short int pol_deg, short int min_harmonic_length,
		   short int max_harmonic_length);
void cut_small_terms(INTERVAL *fn_addr, unsigned long int nterms,
		     double cut_value);
void decode_hom_pol_ind(unsigned long int npos, unsigned short int pol_deg,
			unsigned short int kpol[]);
void decode_trig_ind(unsigned long int npos, short int ktrig[]);
void deriv_action(INTERVAL *fn_addr, unsigned long int nterms,
		  unsigned short int pol_deg, unsigned short int ind_der,
		  INTERVAL *der_fn_addr);
void deriv_angle(INTERVAL *fn_addr, unsigned long int nterms,
		 unsigned short int pol_deg, unsigned short int ind_der,
		 INTERVAL *der_fn_addr);
void determinant(INTERVAL mat[][NDIM], short int n, INTERVAL *det_matr);
char flag_cos_sin(short int ktrig[]);
void lie_deriv_on_expansion(INTERVAL *gen_fn_addr,
			    unsigned short int gen_pol_deg, char flag_transl,
			    INTERVAL *fn_exp_addr[], INTERVAL coef_scal_mul,
			    INTERVAL *target_exp_addr[]);
void move_average_to(INTERVAL *fn_addr, unsigned long int nterms,
		     unsigned short int pol_deg, INTERVAL *aver_fn_addr);
void poisson_index(unsigned short int kpol[][NDIM], short int ktrig[][NDOF],
		   unsigned short int ind_der, INTERVAL coef_res[],
		   unsigned short int kpol_res[], short int ktrig_res[][NDOF]);
void prod_matr_vect(INTERVAL matr[][NDIM], INTERVAL vect[],
		    int n, INTERVAL res_vect[]);
void readd_lie_series_chi2_target_term(unsigned short int ord_gen);
void readd_lie_series_csiq_average_terms(unsigned short int ord_gen);
void scalar_multiply(INTERVAL *fn_addr, unsigned long int nterms,
		     INTERVAL mul_coef);
void scal_prod_vectors(INTERVAL vect1[], INTERVAL vect2[],
		       int n, INTERVAL *res);
void solve_homol_eq(INTERVAL *fn_addr, unsigned long int nterms,
		    double cut_factor, unsigned short int ord_gen,
		    INTERVAL *gen_fn_addr, INTERVAL *target_fn_addr);
void solve_transl_eq(INTERVAL inv_matr_C[][NDIM], INTERVAL *aver_lin_part_addr,
		     INTERVAL *transl_gen_fn_addr);


/* ### */
int main()
{
  unsigned long int i;
  short int n, k[NDIM], ord_gen;
  INTERVAL coef, inv_matr_C[NDIM][NDIM];
  INTERVAL max_l1norm_on_row_inv_matr_C, coef_nondeg_matr_C;
  FILE *ofp;

  /* Calculate some utilities for doing interval arithmetics */
  calc_util_intv_arit();

  /* Fill the indexes tables */
  build_index_pol_table();
  build_index_trig_table();

  /* Allocate the memory cells required by the expansions of all the
     functions are involved in the algorithm. Such functions are
     listed in the GLOBAL pointers variables */
  create_functions();
  /* Define the expansion of the Hamiltonian of the system we are
     considering */
  def_init_ham();

  /* Open the binary file where the upper bounds on some functions
     (useful to complete the computer-assisted proof) will be
     written */
  ofp = fopen("bounds_expl_expans.bin", "wb");
  /* Write on the output binary file the functional norms of the terms
     appearing in the expansion of the Hamiltonian */
  write_norms_ham_terms(ofp);


  /* Algorithm constructing the Kolmogorov's normal form */
  for(ord_gen=1; ord_gen<=KAM_NORMAL_ORDER; ord_gen++) {

    printf("Average of the Hamiltonian part having degree 0 in action:\n");
    printf("%24.16le ,%24.16le\n", ham[0]->dw, ham[0]->up);
    printf("Average of the Hamiltonian part having degree 1 in action:\n");
    printf("%24.16le ,%24.16le\n", ham[1]->dw, ham[1]->up);

    /* Determine the generating function X, which eliminates (part of)
       the perturbing terms having degree 0 in action */
    solve_homol_eq(ham[0], nterms_ham[0], 1.e-2, ord_gen, gen_X, target_X);
    printf("  gen_X related to the step %d :\n", ord_gen);
    print_nz_terms(gen_X, nterms_ham[0], 0, FALSE);
    /* For what concerns the generating function X, write on the output
       file what is needed to complete the computer-assisted proof */
    write_util_estimates_lie_series_X(ofp, ord_gen);
    /* Calculate the new Hamiltonian given by the application of the
       Lie series with generating function X to the old Hamiltonian;
       then write on the output file the functional norms of the terms
       appearing in the expansion of the new Hamiltonian */
    can_transf_ham(gen_X, ord_gen, 0, FALSE);
    write_norms_ham_terms(ofp);

    printf("Average of the Hamiltonian part having degree 0 in action:\n");
    printf("%24.16le ,%24.16le\n", ham[0]->dw, ham[0]->up);
    printf("Average of the Hamiltonian part having degree 1 in action:\n");
    printf("%24.16le ,%24.16le\n", ham[1]->dw, ham[1]->up);

    move_average_to(ham[2], nterms_ham[2], 2, average_ham[2]);
    calc_inv_matr_C(average_ham[2], inv_matr_C);
    /* Determine the generating function csi.q (i.e., that inducing a
       translation of the action), which eliminates the unwanted
       average terms having degree 1 in action */
    move_average_to(ham[1], nterms_ham[1], 1, average_ham[1]);
    solve_transl_eq(inv_matr_C, average_ham[1], gen_csiq);
    printf("  gen_csiq related to the step %d :\n", ord_gen);
    print_nz_terms(gen_csiq, NDIM, 1, TRUE);
    /* For what concerns the generating function csi.q, write on the
       output file what is needed to complete the computer-assisted
       proof */
    write_util_estimates_lie_series_csiq(ofp);
    /* Calculate the new Hamiltonian given by the application of the
       Lie series with generating function csi.q to the old
       Hamiltonian; then write on the output file the functional norms
       of the terms appearing in the expansion of the new Hamiltonian */
    can_transf_ham(gen_csiq, ord_gen, 0, TRUE);
    readd_lie_series_csiq_average_terms(ord_gen);
    write_norms_ham_terms(ofp);

    /* Determine the generating function chi2, which eliminates (part
       of) the perturbing terms having degree 1 in action */
    solve_homol_eq(ham[1], nterms_ham[1], 1.e-3,
		   ord_gen, gen_chi2, target_chi2);
    printf("  gen_chi2 related to the step %d :\n", ord_gen);
    print_nz_terms(gen_chi2, nterms_ham[1], 1, FALSE);
    /* For what concerns the generating function chi2, write on the
       output file what is needed to complete the computer-assisted
       proof */
    write_util_estimates_lie_series_chi2(ofp, ord_gen);
    /* Calculate the new Hamiltonian given by the application of the
       Lie series with generating function chi2 to the old
       Hamiltonian; then write on the output file the functional norms
       of the terms appearing in the expansion of the new Hamiltonian */
    can_transf_ham(gen_chi2, ord_gen, 1, FALSE);
    readd_lie_series_chi2_target_term(ord_gen);
    write_norms_ham_terms(ofp);

  }

  /* Calculate and write on the output file the coefficient related
     to the non-degeneracy of the matrix C (given by the Hessian
     of the average of the quadratic part of the Hamiltonian) */
  move_average_to(ham[2], nterms_ham[2], 2, average_ham[2]);
  calc_inv_matr_C(average_ham[2], inv_matr_C);
  max_l1norm_on_row_inv_matr_C.dw = max_l1norm_on_row_inv_matr_C.up = 0.;
  for(n=0; n<NDIM; n++) {
    short int n_col;
    INTERVAL l1_norm_row;
    l1_norm_row.dw = l1_norm_row.up = 0.;
    for(n_col=0; n_col<NDIM; n_col++)
      add_interval_to( &inv_matr_C[n][n_col] , &l1_norm_row );
    if( max_l1norm_on_row_inv_matr_C.dw < l1_norm_row.dw )
      max_l1norm_on_row_inv_matr_C.dw = l1_norm_row.dw;
    if( max_l1norm_on_row_inv_matr_C.up < l1_norm_row.up )
      max_l1norm_on_row_inv_matr_C.up = l1_norm_row.up;
  }
  inv_interval( &max_l1norm_on_row_inv_matr_C , &coef_nondeg_matr_C );
  fwrite(&coef_nondeg_matr_C.dw, sizeof(double), 1, ofp);

  fclose(ofp);

}


/* ### */
/* create_functions : allocate the arrays where the expansions of the
                      functions handled by the program are
                      stored. Moreover, it determines the number of
                      terms for all the functions */
void create_functions()
{
  unsigned short int l;

  nterms_ham[0] = Npol_table[NDIM][0] * Ntrig_table[NDOF][MAXMODANG];
  for(l=1; l<=MAXPOLDEG; l++)
    nterms_ham[l] = (Npol_table[NDIM][l] - Npol_table[NDIM][l-1]) *
                   Ntrig_table[NDOF][MAXMODANG];
  for(l=0; l<=MAXPOLDEG; l++) {
    ham[l] = (INTERVAL *) calloc(nterms_ham[l], sizeof(INTERVAL));
    old_ham[l] = (INTERVAL *) calloc(nterms_ham[l], sizeof(INTERVAL));
    tmp_ham[0][l] = (INTERVAL *) calloc(nterms_ham[l], sizeof(INTERVAL));
    tmp_ham[1][l] = (INTERVAL *) calloc(nterms_ham[l], sizeof(INTERVAL));
    if( ham[l] == NULL || old_ham[l] == NULL ||
        tmp_ham[0][l] == NULL || tmp_ham[1][l] == NULL ) {
      printf("The requested size of the functions cannot be allocated!!!\n");
      exit(11);
    }
  }

  gen_X = (INTERVAL *) calloc(nterms_ham[0], sizeof(INTERVAL));
  gen_csiq = (INTERVAL *) calloc(NDIM, sizeof(INTERVAL));
  gen_chi2 = (INTERVAL *) calloc(nterms_ham[1], sizeof(INTERVAL));
  if(gen_X == NULL || gen_csiq == NULL || gen_chi2 == NULL) {
    printf("The requested size of the functions cannot be allocated!!!\n");
    exit(11);
  }

  target_X = (INTERVAL *) calloc(nterms_ham[0], sizeof(INTERVAL));
  target_csiq = (INTERVAL *) calloc(NDIM, sizeof(INTERVAL));
  target_chi2 = (INTERVAL *) calloc(nterms_ham[1], sizeof(INTERVAL));
  if(target_X == NULL || target_csiq == NULL || target_chi2 == NULL) {
    printf("The requested size of the functions cannot be allocated!!!\n");
    exit(11);
  }

  nterms_aver_ham[0] = Npol_table[NDIM][0];
  for(l=1; l<=2; l++)
    nterms_aver_ham[l] = Npol_table[NDIM][l] - Npol_table[NDIM][l-1];
  for(l=0; l<=2; l++) {
    average_ham[l] = (INTERVAL *) calloc(nterms_aver_ham[l], sizeof(INTERVAL));
    if(average_ham[l] == NULL) {
      printf("The requested size of the functions cannot be allocated!!!\n");
      exit(11);
    }
  }

}


/* ### */
/* def_init_ham : fill the expansion of the Hamiltonian with
                  the initial definition */
void def_init_ham()
{
  unsigned short int n, kpol[NDIM];
  short int ktrig[NDOF];
  unsigned long int i, pol_npos, trig_npos;
  INTERVAL tmp1_intv, tmp2_intv, coef;

  /* Determine the index corresponding to the Fourier harmonic
     equal to zero */
  for(n=0; n<NDOF; n++)
    ktrig[n] = 0;
  trig_npos = code_trig_ind(ktrig);

  /* Write the expansion of the (integrable) quadratic part of the
     Hamiltonian */
  for(n=0; n<NDIM; n++)
    kpol[n] = 0;
  cv_real_to_interval(0.5, &coef);
  for(n=0; n<NDIM; n++) {
    kpol[n] += 2;
    pol_npos = code_hom_pol_ind(kpol);
    i = pol_npos * Ntrig_table[NDOF][MAXMODANG] + trig_npos;
    copy_interval( &coef , ham[2] + i );
    kpol[n] = 0;
  }
    
  /* Write the expansion of the perturbing part of the Hamiltonian
     depending on the angles only */
  pol_npos = code_hom_pol_ind(kpol);
  cv_real_to_interval(EPSILON, &coef);
  ktrig[0] = 1;
  trig_npos = code_trig_ind(ktrig);
  i = pol_npos * Ntrig_table[NDOF][MAXMODANG] + trig_npos;
  /* The writing of the first angular term is concluded by the next
     instruction */
  copy_interval( &coef , ham[0] + i );
  ktrig[NDOF-1] = -1;
  trig_npos = code_trig_ind(ktrig);
  i = pol_npos * Ntrig_table[NDOF][MAXMODANG] + trig_npos;
  /* The writing of the second angular term is concluded by the next
     instruction */
  copy_interval( &coef , ham[0] + i );

  /* Define the frequency vector omega related to the invariant torus
     we are interested in */
  cv_real_to_interval( sqrt(5.) , &tmp1_intv );
  cv_real_to_interval( -1. , &tmp2_intv );
  add_interval_to( &tmp2_intv , &tmp1_intv );
  cv_real_to_interval( 0.5 , &tmp2_intv );
  mul_interval( &tmp1_intv , &tmp2_intv , omega );
  cv_real_to_interval( 1. , omega + 1 );

}


/* ### */
/* funct_norm : evaluate the functional norm of a part of a function
                (the expansion of which starts from the address
                fn_addr; the number of terms of the expansion is equal
                to nterms and the polynomial degree with respect to
                the action is pol_deg).
		The contribution of a term to the functional norm is
		considered if and only if the corresponding Fourier
		harmonic has a length not smaller than min_harmonic_length
		and not exceeding max_harmonic_length; moreover, the
		constant terms are always disregarded */
void funct_norm(fn_addr, nterms, pol_deg, min_harmonic_length,
		max_harmonic_length, norm)
     INTERVAL *fn_addr;
     unsigned long int nterms;
     unsigned short int pol_deg;
     short int min_harmonic_length, max_harmonic_length;
     INTERVAL *norm;
{
  unsigned short int n, harmonic_length;
  short int ktrig[NDOF];
  unsigned long int i, trig_npos;
  INTERVAL abs_coef;
  INTERVAL *term_addr;

  norm->dw = norm->up = 0.;  

  for(i=0; i<nterms; i++) {

    term_addr = fn_addr + i;
    if( test_intv_different_zero(*term_addr) == TRUE ) {

      trig_npos = i % Ntrig_table[NDOF][MAXMODANG];
      decode_trig_ind(trig_npos, ktrig);
      harmonic_length = 0;
      for(n=0; n<NDOF; n++)
	harmonic_length += ABS(ktrig[n]);

      if( min_harmonic_length <= harmonic_length && 
	  harmonic_length <= max_harmonic_length) {
	if( pol_deg > 0 || harmonic_length > 0 ) {
	  abs_interval_to_interval(term_addr, &abs_coef);
	  add_interval_to(&abs_coef, norm);
	}
      }

    }

  }

}


/* ### */
/* print_nz_terms : print all the non zero terms of a function.
                    REMARK: while print_nz_terms will probably never
                    called in the final version of the program, it is
                    very useful at the debugging stage */
void print_nz_terms(fn_addr, nterms, pol_deg, flag_average)
     INTERVAL *fn_addr;
     unsigned long int nterms;
     unsigned short int pol_deg;
     char flag_average;
{
  unsigned short int n, kpol[NDIM];
  short int ktrig[NDOF];
  unsigned long int i, pol_npos, trig_npos;
  INTERVAL *term_addr;

  for(i=0; i<nterms; i++) {
    term_addr = fn_addr + i;
    if( test_intv_different_zero(*term_addr) == TRUE ) {
      if(flag_average == TRUE) {
	decode_hom_pol_ind(i, pol_deg, kpol);
	printf("(");
	for(n=0; n<NDIM-1; n++)
	  printf("%3d,", kpol[n]);
	printf("%3d) -> ", kpol[NDIM-1]);
      }
      else {
	pol_npos = i / Ntrig_table[NDOF][MAXMODANG];
	decode_hom_pol_ind(pol_npos, pol_deg, kpol);
	printf("(");
	for(n=0; n<NDIM; n++)
	  printf("%3d,", kpol[n]);
	trig_npos = i % Ntrig_table[NDOF][MAXMODANG];
	decode_trig_ind(trig_npos, ktrig);
	for(n=0; n<NDOF-1; n++)
	  printf("%4d,", ktrig[n]);
	printf("%4d) -> ", ktrig[NDOF-1]);
      }
      printf("%24.16le ,%24.16le\n", term_addr->dw, term_addr->up);
    }
  }

}


/* ### */
void write_norms_ham_terms(ofp)
     FILE *ofp;
{
  unsigned short int j, l;
  double norm_up;
  INTERVAL norm;

 for(l=0; l<=MAXPOLDEG; l++) {
   for(j=(l==0); j<=KAM_NORMAL_ORDER; j++) {
     funct_norm(ham[l], nterms_ham[l], l, (j-1)*K+1, j*K, &norm);
     norm_up = norm.up;
     fwrite(&norm_up, sizeof(double), 1, ofp);
   }
 }

}


/* ### */
void write_util_estimates_lie_series_chi2(ofp, ord_gen)
     FILE *ofp;
     unsigned short int ord_gen;
{
  unsigned short int i, n;
  double max_norm_der_chi2;
  INTERVAL norm_der_chi2;

  for(i=0; i<2; i++) {

    max_norm_der_chi2 = 0.;

    for(n=0; n<NDIM; n++) {
      clean_function(tmp_ham[0][1-i], nterms_ham[1-i]);
      if(i == 0)
	deriv_angle(gen_chi2, nterms_ham[1], 1, n, tmp_ham[0][1]);
      else
	deriv_action(gen_chi2, nterms_ham[1], 1, n, tmp_ham[0][0]);
      funct_norm(tmp_ham[0][1-i], nterms_ham[1-i], 1-i,
		 0, ord_gen*K, &norm_der_chi2);
      if(max_norm_der_chi2 < norm_der_chi2.up)
	max_norm_der_chi2 = norm_der_chi2.up;
    }

    fwrite(&max_norm_der_chi2, sizeof(double), 1, ofp);

  }

}


/* ### */
void write_util_estimates_lie_series_csiq(ofp)
     FILE *ofp;
{
  unsigned short int n;
  double max_csi, max_norm_der_X=0.;
  INTERVAL abs_comp_csi;

  fwrite(&max_norm_der_X, sizeof(double), 1, ofp);

  max_csi = 0.;

  for(n=0; n<NDIM; n++) {
    abs_interval_to_interval(gen_csiq + n, &abs_comp_csi);
    if(max_csi < abs_comp_csi.up)
      max_csi = abs_comp_csi.up;
  }
  fwrite(&max_csi, sizeof(double), 1, ofp);

}


/* ### */
void write_util_estimates_lie_series_X(ofp, ord_gen)
     FILE *ofp;
     unsigned short int ord_gen;
{
  unsigned short int n;
  double max_norm_der_X, max_csi=0.;
  INTERVAL norm_der_X;

  max_norm_der_X = 0.;

  for(n=0; n<NDIM; n++) {
    clean_function(tmp_ham[0][0], nterms_ham[0]);
    deriv_angle(gen_X, nterms_ham[0], 0, n, tmp_ham[0][0]);
    funct_norm(tmp_ham[0][0], nterms_ham[0], 0, 0, ord_gen*K, &norm_der_X);
    if(max_norm_der_X < norm_der_X.up)
      max_norm_der_X = norm_der_X.up;
  }

  fwrite(&max_norm_der_X, sizeof(double), 1, ofp);
  fwrite(&max_csi, sizeof(double), 1, ofp);

}


/* ### */
/* add_fn_to : add the first function to the second one (that will
               contain the final result)
               REMARK: the consistency of the number of terms of the
	       summands is not checked! */
void add_fn_to(fn1_addr, nterms, fn2_addr)
     INTERVAL *fn1_addr;
     unsigned long int nterms;
     INTERVAL *fn2_addr;
{
  unsigned long int i;

  for(i=0; i<nterms; i++)
    add_interval_to( (fn1_addr + i) , (fn2_addr + i) );
}


/* ### */
/* add_poisson_to : calculate the Poisson bracket between two
                    functions (the expansions of which start from the
                    addresses fn1_addr and fn2_addr, respectively),
                    multiply the result by the scalar coef_scal_mul
                    and add it to a third function (the expansions of
                    which starts from the address target_fn_addr).
		    REMARK: this function does not check the consistency
		    of the polynomial degree of the function where the
		    Poisson bracket is added with those entering as
		    arguments in the Poisson bracket itself!!! */
void add_poisson_to(fn1_addr, n1_terms, pol1_deg,
		    fn2_addr, n2_terms, pol2_deg,
		    coef_scal_mul, target_fn_addr)
     unsigned long int n1_terms, n2_terms;
     unsigned short int pol1_deg, pol2_deg;
     INTERVAL coef_scal_mul;
     INTERVAL *fn1_addr, *fn2_addr, *target_fn_addr;
{
  unsigned long int j1, j2, pol_npos, trig_npos, res_npos;
  unsigned short int i, kpol[2][NDIM], ind_der, kpol_res[NDIM];
  short int ktrig[2][NDOF], ktrig_res[2][NDOF];
  INTERVAL tmp_coef1, tmp_coef2, coef_res_index[2], coef_fin_res;

  for(j1=0; j1<n1_terms; j1++) {
    if( test_intv_different_zero( *(fn1_addr + j1) ) == TRUE ) {
      /* From the value of the counter j1 reconstruct the
	 corresponding vectors containing the polynomial exponents and
	 the Fourier harmonic */
      pol_npos = j1 / Ntrig_table[NDOF][MAXMODANG];
      decode_hom_pol_ind(pol_npos, pol1_deg, kpol[0]);
      trig_npos = j1 % Ntrig_table[NDOF][MAXMODANG];
      decode_trig_ind(trig_npos, ktrig[0]);
      mul_interval(&coef_scal_mul, fn1_addr + j1, &tmp_coef1);

      for(j2=0; j2<n2_terms; j2++) {
	if( test_intv_different_zero( *(fn2_addr + j2) ) == TRUE ) {
	  /* From the value of the counter j2 reconstruct the
	     corresponding vectors containing the polynomial exponents
	     and the Fourier harmonic */
	  pol_npos = j2 / Ntrig_table[NDOF][MAXMODANG];
	  decode_hom_pol_ind(pol_npos, pol2_deg, kpol[1]);
	  trig_npos = j2 % Ntrig_table[NDOF][MAXMODANG];
	  decode_trig_ind(trig_npos, ktrig[1]);
	  mul_interval(&tmp_coef1, fn2_addr + j2, &tmp_coef2);

	  /* Consider all the derivatives of the Poisson bracket */
	  for(ind_der=0; ind_der<NDIM; ind_der++) {
	    /* Perform the ind_der-th derivative of the Poisson by
	       firstly neglecting the coefficents of both the terms */ 
	    poisson_index(kpol, ktrig, ind_der,
			  coef_res_index, kpol_res, ktrig_res);
	    for(i=0; i<2; i++) {
	      /* From the vectors containing the polynomial exponents
		 and the Fourier harmonics reconstruct the corresponding
		 positions in the vector describing the expansion; then,
		 add the coefficients corresponding to the present terms
		 generated by the Poisson brackets */
	      pol_npos = code_hom_pol_ind(kpol_res);
	      if( test_intv_different_zero( coef_res_index[i] ) == TRUE) {
		mul_interval(&tmp_coef2, coef_res_index+i, &coef_fin_res);
		trig_npos = code_trig_ind(ktrig_res[i]);
		res_npos = pol_npos * Ntrig_table[NDOF][MAXMODANG] + trig_npos;
		add_interval_to(&coef_fin_res , target_fn_addr + res_npos);
	      }
	    }
	  }

	}
      }
    }
  }

}


/* ### */
/* add_poisson_transl_to : calculate the Poisson bracket between a
                           generating function inducing a translation
                           (i.e. of the type csi.q) and another
                           function (the expansions of which start
                           from the addresses gen_transl_addr and
                           fn2_addr, respectively), multiply the
                           result by the scalar coef_scal_mul and add
                           it to a third function (the expansions of
                           which starts from the address
                           target_fn_addr).
			   REMARK: this function does not check the
                           consistency of the polynomial degree of the
                           function where the Poisson bracket is added
                           with those entering as arguments in the
                           Poisson bracket itself!!! */
void add_poisson_transl_to(gen_transl_addr, fn2_addr, n2_terms, pol2_deg,
			   coef_scal_mul, target_fn_addr)
     unsigned long int n2_terms;
     unsigned short int pol2_deg;
     INTERVAL coef_scal_mul;
     INTERVAL *gen_transl_addr, *fn2_addr, *target_fn_addr;
{
  unsigned long int j2, pol_npos, trig_npos, res_npos;
  unsigned short int kpol[NDIM], ind_der;
  INTERVAL tmp_coef1, tmp_coef2, tmp_coef3, mul_factor, fin_coef_add;

  /* Consider all the derivatives of the Poisson bracket */
  for(ind_der=0; ind_der<NDIM; ind_der++) {
    /* Take the ind_der-th component of the translation vector and
       multiply it by the factor coef_scal_mul; store the result in
       the variable tmp_coef1 */
    mul_interval(gen_transl_addr + ind_der, &coef_scal_mul, &tmp_coef1);
    /* If the result of the multiplication above is equal to zero
       there is nothing to do */
    if( test_intv_different_zero( tmp_coef1 ) == TRUE ) {

      /* Consider all the terms of the expansion of the function
	 appearing as the second argument in the Poisson bracket */
      for(j2=0; j2<n2_terms; j2++) {
	copy_interval(fn2_addr + j2, &tmp_coef2);
	if( test_intv_different_zero( tmp_coef2 ) == TRUE ) {
	  /* From the value of the counter j2 reconstruct the
	     corresponding vector containing the polynomial exponents */
	  pol_npos = j2 / Ntrig_table[NDOF][MAXMODANG];
	  decode_hom_pol_ind(pol_npos, pol2_deg, kpol);
	  trig_npos = j2 % Ntrig_table[NDOF][MAXMODANG];

	  /* If the ind_der-th polynomial exponent of the action
	     variables is equal to zero, there is nothing to do */
	  if( kpol[ind_der] != 0 ) {
	    /* Calculate the final numerical coefficient fin_coef_add,
	       that is given by coef_scal_mul times the Poisson
	       bracket between the ind_der-th component of the
	       translation and the j2-th term of the expansion of the
	       function (starting from the address fn2_addr) */ 
	    mul_interval( &tmp_coef1 , &tmp_coef2 , &tmp_coef3 );
	    cv_int_to_interval( kpol[ind_der] , &mul_factor );
	    mul_interval( &tmp_coef3 , &mul_factor , &fin_coef_add );
	    /* The derivative lowers by one the ind_der-th polynomial
	       exponent */
	    kpol[ind_der]--;
	    /* From the vector containing the polynomial exponents and
	       the position of the Fourier harmonics reconstruct the
	       corresponding position in the vector describing the
	       expansion; then, add there the final numerical
	       coefficient fin_coef_add */
	    pol_npos = code_hom_pol_ind(kpol);
	    res_npos = pol_npos * Ntrig_table[NDOF][MAXMODANG] + trig_npos;
	    add_interval_to(&fin_coef_add , target_fn_addr + res_npos);
	  }

	}
      }
    }
  }

}


/* ### */
/* add_scalarmultiply_fn_to : multiply every term of the first function
                              and then add it to the second one (that will
			      contain the final result).
			      REMARK: the consistency of the number of
			      terms of the summands is not checked! */
void add_scalarmultiply_fn_to(fn1_addr, nterms, mul_coef, fn2_addr)
     INTERVAL *fn1_addr;
     unsigned long int nterms;
     INTERVAL mul_coef;
     INTERVAL *fn2_addr;
{
  unsigned long int i;
  INTERVAL tmp1_coef, tmp2_coef;

  for(i=0; i<nterms; i++) {
    copy_interval(fn1_addr + i, &tmp1_coef);
    mul_interval(&tmp1_coef, &mul_coef, &tmp2_coef);
    add_interval_to(&tmp2_coef , fn2_addr + i);
  }
}


/* ### */
void build_index_pol_table()
{
  short int n, s;

  for(s=0; s<=MAXPOLDEG; s++)
    Jpol_table[1][s] = 1;
  for(n=2; n<=NDIM; n++) {
      Jpol_table[n][0] = 1;
      for(s=1; s<=MAXPOLDEG; s++)
	Jpol_table[n][s] = Jpol_table[n][s-1] + Jpol_table[n-1][s];
  }

  for(n=1; n<=NDIM; n++)
    Npol_table[n][0] = Jpol_table[n][0];
  for(n=1; n<=NDIM; n++) {
      for(s=1; s<=MAXPOLDEG; s++)
	Npol_table[n][s] = Npol_table[n][s-1] + Jpol_table[n][s];
  }
}


/* ### */
void build_index_trig_table()
{
  short int n, s;

  Jtrig_table[1][0] = 1;
  for(s=1; s<=MAXMODANG; s++)
    Jtrig_table[1][s] = 2;
  for(n=2; n<=NDOF; n++) {
      Jtrig_table[n][0] = Jtrig_table[n-1][0];
      for(s=1; s<=MAXMODANG; s++)
	Jtrig_table[n][s] = Jtrig_table[n][s-1] +
	                    Jtrig_table[n-1][s-1] + Jtrig_table[n-1][s];
  }

  for(n=1; n<=NDOF; n++)
    Ntrig_table[n][0] = Jtrig_table[n][0];
  for(n=1; n<=NDOF; n++) {
      for(s=1; s<=MAXMODANG; s++)
	Ntrig_table[n][s] = Ntrig_table[n][s-1] + Jtrig_table[n][s];
  }
}


/* ### */
void calc_inv_matr_C(aver_quadr_fn_addr, inv_matr_C)
     INTERVAL *aver_quadr_fn_addr;
     INTERVAL inv_matr_C[NDIM][NDIM];
{
  unsigned short int n, index_dif_0[2], kpol[NDIM];
  char count_index_dif_0;
  unsigned long int pol_npos;
  INTERVAL factor_2, matr_C[NDIM][NDIM];

  cv_real_to_interval(2., &factor_2); 

  for(pol_npos=0; pol_npos<nterms_aver_ham[2]; pol_npos++) {
    decode_hom_pol_ind(pol_npos, 2, kpol);
    count_index_dif_0 = 0;
    for(n=0; n<NDIM; n++) {
      while( kpol[n] > 0 ) {
	index_dif_0[count_index_dif_0] = n;
	count_index_dif_0++;
	kpol[n]--;
      }
    }
    if( index_dif_0[0] == index_dif_0[1] )
      mul_interval( aver_quadr_fn_addr + pol_npos, &factor_2,
		    &(matr_C[index_dif_0[0]][index_dif_0[0]]) );
    else {
      copy_interval( aver_quadr_fn_addr + pol_npos , 
		     &(matr_C[index_dif_0[0]][index_dif_0[1]]) );
      copy_interval( aver_quadr_fn_addr + pol_npos ,
		     &(matr_C[index_dif_0[1]][index_dif_0[0]]) );
    }
  }

  cramer( matr_C , NDIM, inv_matr_C );

}


/* ### */
/* can_transf_ham : calculate the new Hamiltonian given by a canonical
                    transformation; in fact, to the old Hamiltonian it
                    is applied a Lie series with a generating
                    function, which is described by the expansion
                    starting at the address gen_fn_addr (and it is of
                    order ord_gen, it has polynomial degree
                    gen_pol_deg, while the flag flag_transl signals if
                    it is a translation on the actions or not).
                    REMARK: recall that the expansions of the
                    homogeneous polynomials (i.e. having the same
                    degree in actions) making part of the Hamiltonian
                    start from the addresses stored in the vector
                    ham, which is a GLOBAL variables vector!
		    REMARK: before the calling of the present
		    function, the sum of the expansions starting from
		    the addresses stored in the vector ham, describes
		    the OLD Hamiltonian. After the calling, the same
		    sum of expansions gives the NEW Hamiltonian. */
void can_transf_ham(gen_fn_addr, ord_gen, gen_pol_deg, flag_transl)
     INTERVAL *gen_fn_addr;
     unsigned short int ord_gen, gen_pol_deg;
     char flag_transl;
{
  unsigned short int n_calc_terms_lie_series, i, j, l, tmp_flag;
  INTERVAL tmp_intv, coef_scal_mul;

  for(l=0; l<=MAXPOLDEG; l++) {
    clean_function(old_ham[l], nterms_ham[l]);
    add_fn_to(ham[l], nterms_ham[l], old_ham[l]);
  }

  for(i=0; i<=KAM_NORMAL_ORDER; i++) {
 
    n_calc_terms_lie_series = (KAM_NORMAL_ORDER - i) / ord_gen;

    for(l=0; l<=MAXPOLDEG; l++) {
      clean_function(tmp_ham[0][l], nterms_ham[l]);
      add_fn_to(old_ham[l], nterms_ham[l], tmp_ham[0][l]);
      cut_harmonics( tmp_ham[0][l], nterms_ham[l], l, (i - 1) * K + 1, i * K );
    }

    for(j=1; j<=n_calc_terms_lie_series; j++) {

      tmp_flag = j % 2;
      for(l=0; l<=MAXPOLDEG; l++)
	clean_function(tmp_ham[tmp_flag][l], nterms_ham[l]);

      cv_int_to_interval(j, &tmp_intv);
      inv_interval(&tmp_intv, &coef_scal_mul);
      lie_deriv_on_expansion(gen_fn_addr, gen_pol_deg, flag_transl,
			     tmp_ham[1-tmp_flag], coef_scal_mul,
			     tmp_ham[tmp_flag]);

      for(l=0; l<=MAXPOLDEG; l++)
	add_fn_to(tmp_ham[tmp_flag][l], nterms_ham[l], ham[l]);
      
    }

  }

}


/* ### */
/* clean_function : put all the first nterms terms of a function equal
                    to zero */
void clean_function(fn_addr, nterms)
     INTERVAL *fn_addr;
     unsigned long int nterms;
{
  unsigned long int i;
  INTERVAL zero_coef;

  zero_coef.dw = zero_coef.up = 0.;
  for(i=0; i<nterms; i++)
    copy_interval(&zero_coef, fn_addr+i);
}


/* ### */
/* code_hom_pol_ind : from the vector of polynomial indexes kpol
                      determine the corresponding position number (in
                      the ordered set of the indexes of the
                      homogeneous polynomials) */
unsigned long int code_hom_pol_ind(kpol)
     unsigned short int kpol[NDIM];
{
  unsigned long int npos=0;
  unsigned short int  n, pol_deg=0, partial_length;

  for(n=0; n<NDIM; n++)
    pol_deg += kpol[n];
  if(pol_deg == 0)
    return(npos);

  partial_length = pol_deg;
  for(n=0; n<NDIM; n++) {
    if(partial_length == 0)
      break;
    npos += Npol_table[NDIM-n][partial_length-1];
    partial_length -= kpol[n];
  }
  npos -= Npol_table[NDIM][pol_deg-1];
  return(npos);

}


/* ### */
/* code_trig_ind : from the vector of trigonometric indexes ktrig
                   determine the corresponding position number (in the
                   ordered set of the trigonometric indexes) */
unsigned long int code_trig_ind(ktrig)
     short int ktrig[NDOF];
{
  unsigned long int npos=0;
  unsigned short int  n, partial_length=0;

  for(n=0; n<NDOF; n++)
    partial_length += ABS(ktrig[n]);
  for(n=0; n<NDOF; n++) {
    if(partial_length == 0)
      break;
    npos += Ntrig_table[NDOF-n][partial_length-1];
    if((ABS(ktrig[n])) == partial_length) {
      if(ktrig[n] == -partial_length)
	npos += 1;	
    }
    else if(ktrig[n] >= 0)
      npos += Ntrig_table[NDOF-n-1][partial_length-(ABS(ktrig[n]))-1];
    else
      npos += Ntrig_table[NDOF-n-1][partial_length-(ABS(ktrig[n]))];
    partial_length -= ABS(ktrig[n]);
  }
  return(npos);

}


/* ### */
/* cramer : calculate the inverse of the matrix of a nxn matrix (with
            n<=NDIM), by using the Cramer's method */
void cramer(A, n, inv_A)
     INTERVAL A[][NDIM], inv_A[][NDIM];
     short int n;
{
  short int i, j, l, m;
  INTERVAL det_A, b[NDIM], tmp_matr[NDIM][NDIM], det_tmp_matr;
 
  determinant(A, n, &det_A);

  for(i=0; i<n; i++)
    b[i].dw = b[i].up = 0.;

  for(m=0; m<n; m++) {
    cv_int_to_interval( 1 , &(b[m]) );
    for(i=0; i<n; i++) {
      for(j=0; j<n; j++) {
        for(l=0; l<n; l++) {
          if(i == l)
            copy_interval( &(b[j]) , &(tmp_matr[j][l]) );
          else
            copy_interval( &(A[j][l]) , &(tmp_matr[j][l]) );
        }
      }
      determinant( tmp_matr , n , &det_tmp_matr );
      div_interval( &det_tmp_matr , &det_A, &(inv_A[m][i]) );
    }
    b[m].dw = b[m].up = 0.;
  }

}


/* ### */
/* cut_0row_icolumn : orderedly store in the (n-1)x(n-1) matrix
                      reduced_matr all the elements appearing in the
                      nxn matrix matr, except those staying in the
                      0-th row or in the i-th column; n must be not
                      greater than NDIM */
void cut_0row_icolumn(matr, n, i, reduced_matr)
     short int n, i;
     INTERVAL matr[][NDIM], reduced_matr[][NDIM];
{
  short int j, l;

  for(j=1;j<n;j++) {
    for(l=0;l<n;l++) {
      if(l!=i)
        reduced_matr[j-1][l-(l>i)] = matr[j][l];
    }
  }
}


/* ### */
/* cut_harmonics : put equal to zero all the coefficients of a
                   function that are corresponding to harmonics the
                   length (i.e. the l1-norm) of which is smaller than
                   min_harmonic_length or greater than the
                   max_harmonic_length */
void cut_harmonics(fn_addr, nterms, pol_deg,
		   min_harmonic_length, max_harmonic_length)
     INTERVAL *fn_addr;
     unsigned long int nterms;
     unsigned short int pol_deg;
     short int min_harmonic_length, max_harmonic_length;
{
  unsigned short int n, kpol[NDIM], harmonic_length;
  short int ktrig[NDOF];
  unsigned long int i, pol_npos, trig_npos;
  INTERVAL zero_coef;
  INTERVAL *term_addr;

  zero_coef.dw = zero_coef.up = 0.;

  for(i=0; i<nterms; i++) {
    term_addr = fn_addr + i;
    if( test_intv_different_zero(*term_addr) == TRUE ) {
      pol_npos = i / Ntrig_table[NDOF][MAXMODANG];
      decode_hom_pol_ind(pol_npos, pol_deg, kpol);
      trig_npos = i % Ntrig_table[NDOF][MAXMODANG];
      decode_trig_ind(trig_npos, ktrig);
      harmonic_length = 0;
      for(n=0; n<NDOF; n++)
	harmonic_length += ABS(ktrig[n]);
      if( harmonic_length < min_harmonic_length ||
	  harmonic_length > max_harmonic_length )
	copy_interval(&zero_coef, term_addr);
    }
  }
}


/* ### */
/* cut_small_terms : put all the terms, such that their absolute value
                     is smaller than cut_value, equal to zero */
void cut_small_terms(fn_addr, nterms, cut_value)
     INTERVAL *fn_addr;
     unsigned long int nterms;
     double cut_value;
{
  unsigned long int i;
  INTERVAL zero_coef, abs_coef;
  INTERVAL *term_addr;

  zero_coef.dw = zero_coef.up = 0.;

  for(i=0; i<nterms; i++) {
    term_addr = fn_addr + i;
    if( test_intv_different_zero(*term_addr) == TRUE ) {
      abs_interval_to_interval(term_addr, &abs_coef);
      if(abs_coef.up < cut_value)
	copy_interval(&zero_coef, term_addr);
    }
  }

}


/* ### */
/* decode_hom_pol_ind : from the position number npos (in the ordered
                        set of the indexes of the homogeneous
                        polynomials) and the polynomial degree pol_deg
                        determine the corresponding vector of
                        polynomial indexes kpol */
void decode_hom_pol_ind(npos, pol_deg, kpol)
     unsigned long int npos;
     unsigned short int pol_deg, kpol[NDIM];
{
  unsigned long int tail_ind;
  unsigned short int  n, partial_length[NDIM];

  partial_length[0] = pol_deg;
  tail_ind = npos;
  for(n=1; n<NDIM; n++) {
    if(tail_ind == 0)
      partial_length[n] = 0;
    else {
      partial_length[n] = 1;
      while(tail_ind >= Npol_table[NDIM-n][partial_length[n]])
	partial_length[n]++;
      tail_ind -= Npol_table[NDIM-n][partial_length[n]-1];
    }
  }
  kpol[NDIM-1] = partial_length[NDIM-1];
  for(n=0; n<NDIM-1; n++)
    kpol[n] = partial_length[n] - partial_length[n+1];

}


/* ### */
/* decode_trig_ind : from the position number npos (in the ordered set
                     of the trigonometric indexes) determine the
                     corresponding vector of trigonometric indexes
                     ktrig */
void decode_trig_ind(npos, ktrig)
     unsigned long int npos;
     short int ktrig[NDOF];
{
  unsigned long int tail_ind;
  unsigned short int  n, partial_length, partial_length_next;

  tail_ind = npos;
  for(n=0; n<NDOF; n++) {
    if(tail_ind == 0) 
      ktrig[n] = 0;
    else {
      partial_length = 1;
      while(tail_ind >= Ntrig_table[NDOF-n][partial_length])
	partial_length++;
      tail_ind -= Ntrig_table[NDOF-n][partial_length-1];
      if(tail_ind == 0)
	ktrig[n] = partial_length;
      else if(tail_ind == 1) {
	ktrig[n] = -partial_length;
	tail_ind = 0;
      }
      else {
	if(n == NDOF-1) {
	  if(tail_ind % 2 == 0)
	    ktrig[n] = tail_ind / 2;
	  else
	    ktrig[n] = (tail_ind - 1) / 2;
	}
	else {
	  partial_length_next = 1; 
	  while(tail_ind >= 2 * Ntrig_table[NDOF-n-1][partial_length_next])
	    partial_length_next++;
	  tail_ind -= 2 * Ntrig_table[NDOF-n-1][partial_length_next-1];
	  ktrig[n] = partial_length - partial_length_next;
	  if(tail_ind >= Jtrig_table[NDOF-n-1][partial_length_next]) {
	    ktrig[n] *= -1;
	    tail_ind -= Jtrig_table[NDOF-n-1][partial_length_next];
	  }
	  tail_ind += Ntrig_table[NDOF-n-1][partial_length_next-1];
	}
      }
    }
  }
}


/* ### */
/* deriv_action : calculate the derivative with respect to the
                  ind_der-th action variable of a function (whose the
                  expansion starts from the address fn_addr and its
                  number of terms is equal to nterms ) having
                  polynomial degree pol_deg in the action variables;
		  the result is stored in an expansion starting from
                  the address der_fn_addr.
		  REMARK: this function does not check the consistency
		  of the polynomial degree of the function where the
		  derivative is stored!!! */
void deriv_action(fn_addr, nterms, pol_deg, ind_der, der_fn_addr)
     INTERVAL *fn_addr;
     unsigned long int nterms;
     unsigned short int pol_deg, ind_der;
     INTERVAL *der_fn_addr;
{
  unsigned short int kpol[NDIM];
  unsigned long int i, j, pol_npos, trig_npos;
  INTERVAL mul_factor;
  INTERVAL *term_addr;

  for(i=0; i<nterms; i++) {
    term_addr = fn_addr + i;
    if( test_intv_different_zero(*term_addr) == TRUE ) {
      pol_npos = i / Ntrig_table[NDOF][MAXMODANG];
      decode_hom_pol_ind(pol_npos, pol_deg, kpol);
      trig_npos = i % Ntrig_table[NDOF][MAXMODANG];
      if( kpol[ind_der] != 0 ) {
	cv_int_to_interval( kpol[ind_der] , &mul_factor );
	kpol[ind_der]--;
	pol_npos = code_hom_pol_ind(kpol);
	j = pol_npos * Ntrig_table[NDOF][MAXMODANG] + trig_npos;
	mul_interval( term_addr , &mul_factor , der_fn_addr + j );
      }
    }
  }

}


/* ### */
/* deriv_angle : calculate the derivative with respect to the
                 ind_der-th angle variable of a function (whose the
                 expansion starts from the address fn_addr and its
                 number of terms is equal to nterms ) having
                 polynomial degree pol_deg in the action variables;
                 the result is stored in an expansion starting from
                 the address der_fn_addr.
		 REMARK: this function does not check the consistency
		 of the polynomial degree of the function where the
		 derivative is stored!!! */
void deriv_angle(fn_addr, nterms, pol_deg, ind_der, der_fn_addr)
     INTERVAL *fn_addr;
     unsigned long int nterms;
     unsigned short int pol_deg, ind_der;
     INTERVAL *der_fn_addr;
{
  unsigned short int n;
  short int ktrig[NDOF];
  unsigned long int i, j, pure_pol_npos, trig_npos;
  INTERVAL mul_factor;
  INTERVAL *term_addr;

  for(i=0; i<nterms; i++) {
    term_addr = fn_addr + i;
    if( test_intv_different_zero(*term_addr) == TRUE ) {
      trig_npos = i % Ntrig_table[NDOF][MAXMODANG];
      pure_pol_npos = i - trig_npos;
      decode_trig_ind(trig_npos, ktrig);
      if( ktrig[ind_der] != 0 ) {
	cv_int_to_interval( ktrig[ind_der] , &mul_factor );
	for(n=0; n<NDOF; n++)
	  ktrig[n] *= -1;
	j = pure_pol_npos + code_trig_ind(ktrig);
	mul_interval( term_addr , &mul_factor , der_fn_addr + j );
      }
    }
  }

}


/* ### */
/* determinant : calculate the determinant of a nxn matrix (with
                 n<=NDIM), by using the Laplace's recursive method */
void determinant(matr, n, det_matr)
     INTERVAL matr[][NDIM], *det_matr;
     short int n;
{
  short int i;
  INTERVAL res, tmp_matr[NDIM][NDIM], minor;
  INTERVAL tmp1_intv, tmp2_intv, tmp3_intv;

  res.dw = res.up = 0.;

  if(n > 1) {
    for(i=0; i<n; i++) {
      cut_0row_icolumn(matr, n, i, tmp_matr);
      cv_int_to_interval( 1 - 2 * (i % 2) , &tmp1_intv );
      mul_interval( &tmp1_intv , &(matr[0][i]) , &tmp2_intv );
      determinant( tmp_matr , n-1 , &minor );
      mul_interval( &tmp2_intv , &minor , &tmp3_intv );
      add_interval_to(&tmp3_intv, &res);
    }
  }
  else {
    copy_interval( &(matr[0][0]) , &res );
  }

  copy_interval(&res , det_matr);

}



/* ### */
/* flag_cos_sin : if the vector ktrig corresponds to a cosine, return 1;
                  otherwise, if the vector ktrig corresponds to a sine
                  (i.e.  the first non zero component is negative),
                  return -1 */
char flag_cos_sin(ktrig)
     short int ktrig[NDOF];
{
  unsigned short int j;  
  char flag_res;

  flag_res = 0;
  for(j=0; j<NDOF; j++)
    if(ktrig[j] > 0) {
      flag_res = 1;
      break;
    }
    else if(ktrig[j] < 0) {
      flag_res = -1;
      break;
    }

  if(flag_res == 0)
    flag_res = 1;

  return(flag_res);

}


/* ### */
/* lie_deriv_on_expansion : calculate the Lie derivative with respect
                            to a generating function (the expansion of
                            which starts from the address gen_fn_addr)
                            having polynomial degree gen_pol_deg in
                            action variables. The Lie derivative is
                            applied to a function expanded with
                            respect to the polynomial degree of the
                            action variables (the expansions of each
                            sub-function start from the addresses
                            stored in the vector fn_exp_addr).
			    The result of the Lie derivative is
			    multiplied by the scalar factor
			    coef_scal_mul and it is (orderedly) added
			    to another function expanded with respect
			    to the polynomial degree of the action
			    variables (here, the expansions of each
			    sub-function start from the addresses
			    stored in the vector target_exp_addr).
			    The case of the generating function
			    inducing a translation (i.e. of the type
			    csi.q) is handled in a suitable way; this
			    case is announced by the variable
			    flag_transl, when it is equal to TRUE */
void lie_deriv_on_expansion(gen_fn_addr, gen_pol_deg, flag_transl,
			    fn_exp_addr, coef_scal_mul, target_exp_addr)
     unsigned short int gen_pol_deg;
     char flag_transl;
     INTERVAL coef_scal_mul;
     INTERVAL *gen_fn_addr, *fn_exp_addr[], *target_exp_addr[];
{
  unsigned short int l;
  unsigned long int gen_nterms;

  if(flag_transl == TRUE)
    gen_nterms = NDIM;
  else
    if(gen_pol_deg == 0)
      gen_nterms = nterms_ham[0];
    else
      gen_nterms = nterms_ham[1];

  if(flag_transl == FALSE)
    for(l=1-gen_pol_deg; l<=MAXPOLDEG; l++)
      add_poisson_to( gen_fn_addr, gen_nterms, gen_pol_deg,
		      fn_exp_addr[l], nterms_ham[l], l,
		      coef_scal_mul, target_exp_addr[l-1+gen_pol_deg] );
  else
    for(l=1; l<=MAXPOLDEG; l++)
      add_poisson_transl_to( gen_fn_addr, fn_exp_addr[l], nterms_ham[l],
			     l, coef_scal_mul, target_exp_addr[l-1] );

}


/* ### */
/* move_average_to : move all the average terms (i.e. those with the
                     Fourier harmonic equal to zero) from the
                     expansion starting from fn_addr to that starting
                     from aver_fn_addr */
void move_average_to(fn_addr, nterms, pol_deg, aver_fn_addr)
     INTERVAL *fn_addr;
     unsigned long int nterms;
     unsigned short int pol_deg;
     INTERVAL *aver_fn_addr;
{
  unsigned short int n, kpol[NDIM];
  short int ktrig[NDOF];
  unsigned long int i, pol_npos, trig_npos;
  char flag_average;
  INTERVAL zero_coef;

  zero_coef.dw = zero_coef.up = 0.;

  for(i=0; i<nterms; i++) {
    pol_npos = i / Ntrig_table[NDOF][MAXMODANG];
    decode_hom_pol_ind(pol_npos, pol_deg, kpol);
    trig_npos = i % Ntrig_table[NDOF][MAXMODANG];
    decode_trig_ind(trig_npos, ktrig);
    flag_average = TRUE;
    for(n=0; n<NDOF; n++)
      if(ktrig[n] != 0) {
	flag_average = FALSE;
	break;
      }
    if(flag_average == TRUE) {
      copy_interval( fn_addr + i , aver_fn_addr + pol_npos );
      copy_interval( &zero_coef, fn_addr + i);
    }
  }

}


/* ### */
/* poisson_index : perform the calculation of the term corresponding
                   to the ind_der-th derivative of the Poisson
                   bracket.  The calculation just involves the indexes
                   corresponding to the polynomial exponents and the
                   Fourier harmonics.
		   The indexes corresponding to the polynomial
		   exponents of the two arguments of the Poisson
		   bracket are stored in the array kpol;
		   while the indexes corresponding to the Fourier
		   harmonics are stored in the array ktrig.
		   For what concerns the terms produced by the
		   ind_der-th derivative of the Poisson bracket,
		   the coefficients, the polynomial exponents and
		   the (two possible) Fourier harmonics are stored
		   in coef_res, kpol_res and ktrig_res, respectively */
void poisson_index(kpol, ktrig, ind_der, coef_res, kpol_res, ktrig_res)
     unsigned short int kpol[2][NDIM], ind_der, kpol_res[NDIM];
     short int ktrig[2][NDOF], ktrig_res[2][NDOF];
     INTERVAL coef_res[2];
{
  short int i, j, sign_pois;
  short int coef_der[2], coef_add_pois[2][2];
  char flag_zero = TRUE, flag_factors[2], flag_prod, flag_tmp;

  /* In this function, hereafter, sign_pois can assume the values 0 or 1:
       sign_pois = 0 means that we are considering the part of the Poisson
                     bracket with minus sign (i.e. we perform the derivative
		     of the first function with respect to the actions
		     and of the second one with respect to the angles);
       sign_pois = 1 means that we are considering the part of the Poisson
                     bracket with plus sign (i.e. we perform the derivative
		     of the first function with respect to the angles
		     and of the second one with respect to the actions);
  */

  /* Calculate the coefficients due to the factors produced by the
     derivatives */
  for(sign_pois=0; sign_pois<2; sign_pois++) {
    coef_der[sign_pois] = (2 * sign_pois - 1) * kpol[sign_pois][ind_der] *
	                  ktrig[1-sign_pois][ind_der];
    if(coef_der[sign_pois] != 0)
      flag_zero = FALSE;
  }

  /* If both the previously defined coefficients (due to the factors
     produced by the derivatives) are equal to zero, there are not
     many things to do */
  if(flag_zero == TRUE) {
    for(i=0; i<2; i++)
      coef_res[i].dw = coef_res[i].up = 0.;
    for(j=0; j<NDIM; j++)
      kpol_res[j] = 0;
    return;
  }

  /* Calculate the polynomial exponents of the terms produced by the
     ind_der-th derivative of the Poisson bracket */
  for(j=0; j<NDIM; j++)
    kpol_res[j] = kpol[0][j] + kpol[1][j];
  kpol_res[ind_der]--;


  /* Put the flags flag_factors[0], flag_factors[1] and flag_prod
     equal to 1 or -1, according to the harmonic of the first argument
     of the Poisson bracket, the harmonic of the second argument and
     their product correspond to a cosine or a sine, respectively */
  for(i=0; i<2; i++)
    flag_factors[i] = flag_cos_sin(ktrig[i]);
  flag_prod = flag_factors[0] * flag_factors[1];

  /* In this function, hereafter, i can assume the values 0 or 1 and
     it is defined so that
       i = 0 means that we DON'T KNOW whether the corresponding Fourier
             harmonic ktrig_res[0][0], ...,  ktrig_res[0][NDOF-1]
	     is corresponding to a cosine or a sine;
       i = 1 means that we KNOW that the corresponding Fourier
             harmonic ktrig_res[1][0], ...,  ktrig_res[1][NDOF-1]
	     is corresponding to either a cosine or a sine; */

  /* Calculate the Fourier harmonics of the terms produced by the
     ind_der-th derivative of the Poisson bracket.
     REMARK: the sign of these harmonics will be rearranged later */
  for(i=0; i<2; i++)
    for(j=0; j<NDOF; j++)
      ktrig_res[i][j] = ktrig[0][j] + (2*i - 1) * flag_prod * ktrig[1][j];

  /* Calculate the sign of the summands generated by product of a harmonic
     with the derivative (with respect to an angle) of another harmonic */
  for(sign_pois=0; sign_pois<2; sign_pois++)
    for(i=0; i<2; i++)
      coef_add_pois[sign_pois][i] = 1;
  coef_add_pois[(1-flag_factors[1])/2][0] = -1;
  for(sign_pois=0; sign_pois<2; sign_pois++)
    for(i=0; i<2; i++)
      coef_add_pois[sign_pois][i] *= -flag_factors[1-sign_pois];

  /* Since there is a derivative with respect to an angle, if the
     product of the harmonics can be expressed by a sum of two
     cosines, then the terms produced by the Poisson bracket will be
     expressed as a sum of two sines, and vice versa */
  flag_prod *= -1;

  /* We now eventually multiply by -1 all the indexes of a Fourier
     harmonic produced by the Poisson bracket, in order that such a
     Fourier harmonic is a cosine when the arguments of the Poisson
     bracket are one sine and one cosine; instead, the Fourier
     harmonic must be a sine in the remaining cases */
  for(i=0; i<2; i++) {
    /* Put the flag flag_tmp equal to 1 or -1, according to the
       harmonic ktrig_res[i] now corresponds to a cosine or a sine,
       respectively. Put flag_tmp equal to zero if none of the indexes
       of the harmonic ktrig_res[i] is different from zero */
    flag_tmp = 0;
    for(j=0; j<NDOF; j++)
      if(ktrig_res[i][j] > 0) {
	flag_tmp = 1;
	break;
      }
      else if(ktrig_res[i][j] < 0) {
	flag_tmp = -1;
	break;
      }
    if(flag_tmp == 0) {
      /* If all the indexes of the harmonic ktrig_res[i] are equal to
         zero and the Poisson bracket must produce two sines, then put
         the corresponding coefficients equal to zero */
      if(flag_prod == -1)
	for(sign_pois=0; sign_pois<2; sign_pois++)
	  coef_add_pois[sign_pois][i] = 0;
    }
    else if(flag_tmp != flag_prod) {
      /* In the present case the temporary value of the Fourier
         harmonic produced by the Poisson bracket corresponds to a
         sine while it should be a cosine or vice versa: we must
         change the sign to all the indexes of the Fourier harmonic */
      for(j=0; j<NDOF; j++)
	ktrig_res[i][j] *= -1;
      /* If the Fourier harmonic produced by the Poisson bracket must
         correspond to a sine, we have also to change the sign to the
         corresponding coefficients */
      if(flag_prod == -1)
	for(sign_pois=0; sign_pois<2; sign_pois++)
	  coef_add_pois[sign_pois][i] *= -1;
    }
  }


  /* Add up the two coefficents corresponding to the same Fourier
     harmonic, by taking into account also the factors produced by the
     derivatives and the factor 1/2 produced by the trigonometric
     formulas we implicitly used above */
  for(i=0; i<2; i++) {
    cv_real_to_interval(0., coef_res + i);
    for(sign_pois=0; sign_pois<2; sign_pois++) {
      INTERVAL tmp_intv;
      cv_real_to_interval( (coef_der[sign_pois] * coef_add_pois[sign_pois][i])
			   * 0.5 , &tmp_intv );
      add_interval_to(&tmp_intv , coef_res + i);
    }
  }

}


/* ### */
/* prod_matr_vect : calculate the product of the nxn matrix matr
                    by the n-dimensional vector vect (where n<=NDIM);
		    the result is stored in the vector res_vect */
void prod_matr_vect(matr, vect, n, res_vect)
     INTERVAL matr[][NDIM], vect[], res_vect[];
     int n;
{
  int i, j;
  INTERVAL tmp_intv;

  for(i=0;i<n;i++) {
    res_vect[i].dw = res_vect[i].up = 0.;
    for(j=0;j<n;j++) {
      mul_interval( &(matr[i][j]) , &(vect[j]) , &tmp_intv );
      add_interval_to( &tmp_intv , &(res_vect[i]) );
    }
  }

}


/* ### */
/* readd_lie_series_chi2_target_term : readd to the Hamiltonian the
                                       terms due to the application of
                                       the Lie series with generating
                                       function chi2 to the two
                                       following linear terms: omega.p
                                       and the target part eliminated
                                       by the generating function chi2
                                       itself */
void readd_lie_series_chi2_target_term(ord_gen)
     unsigned short int ord_gen;
{
  unsigned short int n_calc_terms_lie_series, i, j, tmp_flag;
  INTERVAL tmp1_intv, tmp2_intv, tmp3_intv, coef_scal_mul;

  for(i=1; i<=ord_gen; i++) {

    n_calc_terms_lie_series = (KAM_NORMAL_ORDER - i) / ord_gen;

    clean_function(tmp_ham[0][1], nterms_ham[1]);
    add_fn_to(target_chi2, nterms_ham[1], tmp_ham[0][1]);
    cut_harmonics( tmp_ham[0][1], nterms_ham[1], 1, (i - 1) * K + 1, i * K );
    cv_real_to_interval(0.5, &coef_scal_mul);

    for(j=1; j<=n_calc_terms_lie_series; j++) {

      tmp_flag = j % 2;
      clean_function(tmp_ham[tmp_flag][1], nterms_ham[1]);

      add_poisson_to( gen_chi2, nterms_ham[1], 1,
		      tmp_ham[1-tmp_flag][1], nterms_ham[1], 1,
		      coef_scal_mul, tmp_ham[tmp_flag][1] );
      cv_int_to_interval(j+1, &tmp1_intv);
      cv_int_to_interval(j*(j+2), &tmp2_intv);
      div_interval(&tmp1_intv, &tmp2_intv, &coef_scal_mul);
      add_fn_to(tmp_ham[tmp_flag][1], nterms_ham[1], ham[1]);

    }

  }

}


/* ### */
/* readd_lie_series_csiq_average_terms : readd to the Hamiltonian the
                                         average of the quadratic term
                                         and a constant term due to
                                         the application of the Lie
                                         series with generating
                                         function inducing a
                                         translation (i.e. of the type
                                         csi.q) to the average of the
                                         quadratic term itself and to
                                         the linear term omega.p */
void readd_lie_series_csiq_average_terms(ord_gen)
     unsigned short int ord_gen;
{
  unsigned short int n, kpol[NDIM];
  short int ktrig[NDOF];
  unsigned long int i, pure_pol_npos, trig_npos;
  INTERVAL tmp1_intv, tmp2_intv, tmp3_intv;

  /* Determine the index corresponding to the Fourier harmonic
     equal to zero */
  for(n=0; n<NDOF; n++)
    ktrig[n] = 0;
  trig_npos = code_trig_ind(ktrig);

  /* Readd the average terms of the quadratic part of the Hamiltonian */
  for(pure_pol_npos=0; pure_pol_npos<nterms_aver_ham[2]; pure_pol_npos++) {
    i = pure_pol_npos * Ntrig_table[NDOF][MAXMODANG] + trig_npos;
    add_interval_to( average_ham[2] + pure_pol_npos , ham[2] + i );
  }

  /* Add the constant term csi . omega
     REMARK: the constant terms have not any influence on the dynamics
     but they contribute to the calculation of the energy level of the
     final invariant torus */
  scal_prod_vectors( gen_csiq , omega , NDIM , &tmp1_intv ); 
  for(n=0; n<NDIM; n++)
    kpol[n] = 0;
  pure_pol_npos = code_hom_pol_ind(kpol);
  i = pure_pol_npos * Ntrig_table[NDOF][MAXMODANG] + trig_npos;
  add_interval_to( &tmp1_intv , ham[0] + i );

  /* Add the constant term
           coef * ( csi . d <h_1> / dp )
     where <.> indicates the average; the derivative with respect to
     p must be read, more exactly, as a gradient with respect to
     all the actions; csi . means the scalar product with the vector
     csi related to the translation; the value of coef is equal to
     1/2 or 1.
     coef can be equal to 1/2, because
           1/2 csi . ( d2 <h_2> / dpdp ) csi
     (where the term between the parentheses means the hessian of the
     average of the quadratic part of the Hamiltonian) is equal to
     -1/2 csi . d <h_1> / dp ; but, according to the truncation rules
     used every time the function cut_harmonics is called, the
     contribution of 1/2 csi . ( d2 <h_2> / dpdp ) csi must be
     considered just when 2 * ord_gen <= KAM_NORMAL_ORDER */
  scal_prod_vectors( gen_csiq , average_ham[1] , NDIM , &tmp1_intv );
  if( 2 * ord_gen <= KAM_NORMAL_ORDER )
    cv_real_to_interval(0.5, &tmp2_intv);
  else
    cv_int_to_interval(1, &tmp2_intv);
  mul_interval(&tmp1_intv, &tmp2_intv, &tmp3_intv);
  add_interval_to( &tmp3_intv , ham[0] + i );

}


/* ### */
/* scal_prod_vectors : calculate the scalar product between two
                       n-dimensional vectors vect1 and vect2
		       the result is stored in res */
void scal_prod_vectors(vect1, vect2, n, res)
     INTERVAL vect1[], vect2[], *res;
     int n;
{
  int j;
  INTERVAL tmp_intv;

  res->dw = res->up = 0.;
  for(j=0;j<n;j++) {
    mul_interval( &(vect1[j]) , &(vect2[j]) , &tmp_intv );
    add_interval_to( &tmp_intv , res );
  }

}


/* ### */
/* scalar_multiply : multiply all the terms of a function by a scalar
                     factor */
void scalar_multiply(fn_addr, nterms, mul_coef)
     INTERVAL *fn_addr;
     unsigned long int nterms;
     INTERVAL mul_coef;
{
  unsigned long int i;
  INTERVAL *fn_coef_addr;
  INTERVAL old_fn_coef;

  for(i=0; i<nterms; i++) {
    fn_coef_addr = fn_addr + i;
    copy_interval(fn_coef_addr, &old_fn_coef);
    mul_interval(&old_fn_coef, &mul_coef, fn_coef_addr);
  }
}


/* ### */
/* solve_homol_eq : solve the homological equation.
                    The expansion of the perturbing function to
		    (partially) eliminate starts from the address
		    fn_addr and it contains nterms.
                    The expansion of the generating function starts
		    from gen_fn_addr and the maximal length of its
		    harmonics is ord_gen*K; moreover, all the terms,
		    whose the corresponding coefficient is smaller
		    than the maximal one multiplied by cut_factor, are
		    put equal to zero.
		    All the terms that will be eliminated from the
		    perturbing function (when the Lie series with such
		    generating function is applied to the Hamiltonian)
		    are moved from the perturbing function to the
		    expansion starting from target_fn_addr */
void solve_homol_eq(fn_addr, nterms, cut_factor,
		    ord_gen, gen_fn_addr, target_fn_addr)
     unsigned long int nterms;
     double cut_factor;
     unsigned short int ord_gen;
     INTERVAL *fn_addr, *gen_fn_addr, *target_fn_addr;
{
  unsigned short int n, harmonic_length;
  short int ktrig[NDOF];
  unsigned long int i, j, pure_pol_npos, trig_npos;
  double max_coef_gen_fn = 0.;
  INTERVAL zero_coef, denom, coef, abs_coef;
  INTERVAL *term_addr;

  zero_coef.dw = zero_coef.up = 0.;

  clean_function(gen_fn_addr, nterms);
  clean_function(target_fn_addr, nterms);

  for(i=0; i<nterms; i++) {
    term_addr = fn_addr + i;
    if( test_intv_different_zero(*term_addr) == TRUE ) {
      trig_npos = i % Ntrig_table[NDOF][MAXMODANG];
      pure_pol_npos = i - trig_npos;
      decode_trig_ind(trig_npos, ktrig);
      harmonic_length = 0;
      for(n=0; n<NDOF; n++)
	harmonic_length += ABS(ktrig[n]);
      if( harmonic_length > 0 && harmonic_length <= ord_gen*K ) {
	copy_interval(&zero_coef, &denom);
	for(n=0; n<NDOF; n++) {
	  INTERVAL tmp1_intv, tmp2_intv;
	  cv_int_to_interval(ktrig[n], &tmp1_intv);
	  mul_interval(&tmp1_intv, omega+n, &tmp2_intv);
	  add_interval_to(&tmp2_intv, &denom);
	  ktrig[n] *= -1;
	}
	div_interval(term_addr, &denom, &coef);
	j = pure_pol_npos + code_trig_ind(ktrig);
	copy_interval( &coef , gen_fn_addr + j );
	abs_interval_to_interval(&coef, &abs_coef);
	if(abs_coef.up > max_coef_gen_fn)
	  max_coef_gen_fn = abs_coef.up;
      }
    }
  }

  cut_small_terms( gen_fn_addr, nterms, cut_factor*max_coef_gen_fn );

  for(j=0; j<nterms; j++) {
    if( test_intv_different_zero( *(gen_fn_addr + j) ) == TRUE ) {
      trig_npos = j % Ntrig_table[NDOF][MAXMODANG];
      pure_pol_npos = j - trig_npos;
      decode_trig_ind(trig_npos, ktrig);
      for(n=0; n<NDOF; n++)
	ktrig[n] *= -1;
      i = pure_pol_npos + code_trig_ind(ktrig);
      copy_interval( fn_addr + i , target_fn_addr + i );
      copy_interval( &zero_coef, fn_addr + i);
    }
  }

}


/* ### */
void solve_transl_eq(inv_matr_C, aver_lin_part_addr, transl_gen_fn_addr)
     INTERVAL inv_matr_C[][NDIM];
     INTERVAL *aver_lin_part_addr, *transl_gen_fn_addr;
{
  INTERVAL factor_minus1;

  cv_int_to_interval(-1, &factor_minus1);
  prod_matr_vect(inv_matr_C, aver_lin_part_addr, NDIM, transl_gen_fn_addr);
  scalar_multiply(transl_gen_fn_addr, NDIM, factor_minus1);

}
